package run

import (
	"context"
	"fmt"
	"os"
	"os/signal"
	"syscall"

	"gitea.dwysokinski.me/twhelp/dcbot/cmd/dcbot/internal"
	"gitea.dwysokinski.me/twhelp/dcbot/internal/bundb"
	"gitea.dwysokinski.me/twhelp/dcbot/internal/discord"
	"gitea.dwysokinski.me/twhelp/dcbot/internal/service"
	"github.com/kelseyhightower/envconfig"
	"github.com/urfave/cli/v2"
	"go.uber.org/zap"
)

const (
	healthyFilePath = "/tmp/healthy"
	healthyFilePerm = 0644
)

func New() *cli.Command {
	return &cli.Command{
		Name:  "run",
		Usage: "Runs the dc bot",
		Action: func(c *cli.Context) error {
			logger := zap.L()

			cfg, err := newBotConfig()
			if err != nil {
				return fmt.Errorf("newBotConfig: %w", err)
			}

			db, err := internal.NewBunDB()
			if err != nil {
				return fmt.Errorf("internal.NewBunDB: %w", err)
			}
			defer func() {
				_ = db.Close()
			}()

			client, err := internal.NewTWHelpClient(c.App.Version)
			if err != nil {
				return fmt.Errorf("internal.NewTWHelpClient: %w", err)
			}

			groupRepo := bundb.NewGroup(db)
			monitorRepo := bundb.NewMonitor(db)

			choiceSvc := service.NewChoice(client)
			groupSvc := service.NewGroup(
				groupRepo,
				client,
				logger,
				cfg.MaxGroupsPerServer,
			)
			monitorSvc := service.NewMonitor(
				monitorRepo,
				groupRepo,
				client,
				logger,
				cfg.MaxMonitorsPerGroup,
			)

			bot, err := discord.NewBot(cfg.Token, groupSvc, monitorSvc, choiceSvc, logger)
			if err != nil {
				return fmt.Errorf("discord.NewBot: %w", err)
			}

			go func() {
				if runErr := bot.Run(); runErr != nil {
					logger.Fatal("bot.Run: " + runErr.Error())
				}
			}()
			defer func() {
				_ = bot.Close()
			}()

			if err = os.WriteFile(healthyFilePath, []byte("healthy"), healthyFilePerm); err != nil {
				return fmt.Errorf("couldn't create file (path=%s): %w", healthyFilePath, err)
			}

			logger.Info("Bot is up and running")

			waitForSignal(c.Context)

			return nil
		},
	}
}

type botConfig struct {
	Token               string `envconfig:"TOKEN" required:"true"`
	MaxGroupsPerServer  int    `envconfig:"MAX_GROUPS_PER_SERVER" default:"10"`
	MaxMonitorsPerGroup int    `envconfig:"MAX_MONITORS_PER_GROUP" default:"10"`
}

func newBotConfig() (botConfig, error) {
	var cfg botConfig
	if err := envconfig.Process("BOT", &cfg); err != nil {
		return botConfig{}, fmt.Errorf("envconfig.Process: %w", err)
	}
	return cfg, nil
}

func waitForSignal(ctx context.Context) {
	ctx, stop := signal.NotifyContext(ctx, os.Interrupt, syscall.SIGTERM)
	defer stop()
	<-ctx.Done()
}
